#include <stdio.h>
#include <stdlib.h>
#include <stdbool.h>
#include <ctype.h>

#include "syntax.h"
#include "extstring.h"

struct SyntaxElement* parse(FILE* file) {
	struct SyntaxElement* syntaxtree = se_create();
	struct SyntaxElement* syntaxtreestart = syntaxtree;
	syntaxtree->type = TOPTREE;
	char symbol;
	bool intree = false;
	bool incomment = false;
	bool instring = false;
	bool intoken = false;
	struct Extstring* exttoken = NULL;
	struct Extstring* extstring = NULL;
	while (1) {
		symbol = fgetc(file);
		if (feof(file))
			break;
		if (incomment) {
			if (symbol == '\n')
				incomment = false;
			continue;
		}
		if (instring) {
			if (symbol == '"') {
				if (extstring == NULL) {
					extstring = malloc(sizeof(struct Extstring));
					extstring->next = NULL;
				}
				instring = false;
				syntaxtree->type = STRING;
				char* string = es_tostring(extstring);
				syntaxtree->content = string;
				es_free(extstring);
				extstring = NULL;
				syntaxtree = se_next(syntaxtree);
			} else {
				if (extstring == NULL) {
					extstring = malloc(sizeof(struct Extstring));
					extstring->symbol = symbol;
					extstring->next = NULL;
					continue;
				}
				es_addsymbol(extstring, symbol);
			}
			continue;
		}
		if (intoken) {
			if (isspace(symbol) || symbol == '(' || symbol == ')') {
				intoken = false;
				syntaxtree->type = TOKEN;
				char* string = es_tostring(exttoken);
				syntaxtree->content = string;
				es_free(exttoken);
				exttoken = NULL;
				syntaxtree = se_next(syntaxtree);
			} else {
				es_addsymbol(exttoken, symbol);
				continue;
			}
		}
		if (isspace(symbol)) 
			continue;
		if (intree) {
			switch (symbol) {
				case '(':
					syntaxtree = se_bottom(syntaxtree);
					break;
				case ')':
					if (syntaxtree->top->type == TOPTREE) {
						intree = false;
						syntaxtree = syntaxtree->top;
						syntaxtree = se_next(syntaxtree);
						syntaxtree->type = NONE;
					} else {
						syntaxtree = syntaxtree->top;
						syntaxtree = se_next(syntaxtree);
						syntaxtree->type = TREE;
					}
					break;
				case ';':
					incomment = true;
					continue;
				case '"':
					instring = true;
					continue;
				default:
					exttoken = malloc(sizeof(struct Extstring));
					exttoken->symbol = symbol;
					exttoken->next = NULL;
					intoken = true;
					break;
			}
			continue;
		} else {
			switch (symbol) {
				case ')':
					fprintf(stderr, "Unbalanced brackets.\n");
					return NULL;
				case ';':
					incomment = true;
					continue;
				case '(':
					intree = true;
					syntaxtree->type = TOPTREE;
					syntaxtree = se_bottom(syntaxtree);
					syntaxtree->type = TREE;
					continue;
				default:
					fprintf(stderr, "Expression outside of brackets.\n");
					return NULL;
			}
		}
	}
	if (intree) {
		fprintf(stderr, "Unexpected EOF.\n");
		return NULL;
	}
	se_clean(syntaxtreestart);
	return syntaxtreestart;
}
